"""
Collection of backends.

Each submodule defines a `backend` dictionary with the backend options. This
dictionary (and all user-defined backends that may be passed to the
`register_backend` class method of clusterjob.Job) must
have the following structure of keys and values:

    'name': str
        Name of the backend.

    'prefix': str
        prefix to be added before each submission options, in the header of the
        job script. E.g. '#SBATCH' for slurm and '#PBS' for PBS/Torque.

    'extension': str
        Default filename extension for job script files.

    'cmd_submit': tuple of (list, callable)
        The first element of the tuple is the shell command that is used for
        submission. It must take the name of the job script file as a
        parameter. The second element of the tuple is a callable that receives
        the shell output from the submission command and returns the job ID
        that the cluster has assigned to the job as a string

    'cmd_status_running': tuple of (list, callable)
        The first element of the tuple is the shell command that yields some
        information about a running job. The placeholder `{job_id}` can be used
        for the cluster-assigned job ID.  The second element of the tuple is a
        callable that receives the shell output from the command and returns
        one of the status codes defined in the `clusterjob.status` module, or
        None if no status can be determined from the output of the command.

    'cmd_cancel': list
        Shell command that cancels the job on the cluster. The placeholder
        `{job_id}` can be used for the cluster-assigned job ID

    'cmd_status_finished': tuple of (list, callable)
        A fallback if `cmd_status_running` is not able to determine a status,
        e.g. because the command defined there does not return any output for
        jobs that have finished.  It will only be called if
        `cmd_status_running` results in None.

    'translate_options': callable
        The callable receives a dictionary of options (from
        clusterjob.Job.options), and must return an array of command line
        options for the backend's submission script. These, together with the
        backend's prefix will be written to the header of the job submission
        script.

    'default_opts': dict
        Default entries for `clusterjob.Job.options` when a Job instance is
        created with the given backend

    'job_vars': dict
        Mapping of replacements that will be applied to the body of the job
        script. The intention is to adjust the name of environment variables to
        the backend, e.g. '$SLURM_JOB_ID' for SLURM vs. '$PBS_JOBID' for
        PBS/Torque. It must define replacements for the following strings:
        'XXX_JOB_ID'      => var containing cluster assigned job ID
        'XXX_WORKDIR'     => var containing submission directory on the cluster
        'XXX_HOST'        => var containing hostname of submission node
        'XXX_JOB_NAME'    => var containing job name
        'XXX_ARRAY_INDEX' => var containing job array/tast index
        'XXX_NODELIST'    => var containing  hostname(s) of job nodes
"""

import slurm

def check_backend(backend, raise_exception=True):
    """
    Return True if the given backend has the correct structure (as compared
    agains the slurm backend)

    Arguments
    ---------

    backend: dict
        Dictionary of backend options
    raise_exceptions: boolean, optional
        If True (default), raise an AssertionError if the backend does not
        match the required structure. Otherwise, return False.
    """
    template = slurm.backend
    try:
        for key in template:
            assert key in backend, "backend is missing mandatory key %s" % key
        test_options = {
            'name': 'testjob',
            'queue': 'testqueue',
            'time': '01:00:00',
            'nodes': 1,
            'threads': 1,
            'mem': 1024,
            'stdout': 'stdout.log',
            'stderr': 'stderr.log',
        }
        try:
            opt_array = backend['translate_options'](test_options)
            for option in opt_array:
                assert str(option) == option
        except Exception as e:
            raise AssertionError("invalid backend %s: %s", backend['name'], e)
        for key in template['job_vars']:
            assert key in backend['job_vars'], \
            "backend does not recognize job variable %s" % key
    except AssertionError:
        if raise_exception:
            raise
        else:
            return False
    return True




"""
SLURM backend
"""
from ..status import PENDING, RUNNING, COMPLETED, CANCELLED, FAILED

opt_replacements = {
    'jobname': ('--job-name',      lambda s: str(s).strip() ),
    'queue'  : ('--partition',     lambda s: str(s).strip() ),
    'time'   : ('--time',          lambda s: str(s).strip() ),
    'nodes'  : ('--nodes',         lambda s: str(s).strip() ),
    'threads': ('--cpus-per-task', lambda s: str(s).strip() ),
    'mem'    : ('--mem',           lambda s: str(s).strip() ),
    'stdout' : ('--output',        lambda s: str(s).strip() ),
    'stderr' : ('--error',         lambda s: str(s).strip() ),
}


def translate_options(options_dict):
    """Translate dictionary of options into array of options for SLURM"""
    opt_array = []
    for (key, val) in options_dict.items():
        if key in opt_replacements:
            slurm_key, converter = opt_replacements[key]
            val = converter(val)
        else:
            slurm_key = key
        if not slurm_key.startswith('-'):
            if len(slurm_key) == 1:
                slurm_key = '-%s' % slurm_key
            else:
                slurm_key = '--%s' % slurm_key
        if val is None:
            continue
        if type(val) is bool:
            if val:
                opt_array.append(slurm_key)
        else:
            if slurm_key.startswith('--'):
                opt_array.append('%s=%s' % (slurm_key, str(val)))
            else:
                opt_array.append('%s %s' % (slurm_key, str(val)))
    return opt_array


def get_job_id(response):
    """Return the job id from the response of the sbatch command"""
    import re
    match = re.search('Submitted batch job (\d+)\s*$', response)
    if match:
        return match.group(1)
    else:
        return None


def get_job_status(response):
    """Return job job status code from the response of the sacct/squeue
    command"""
    slurm_status_mapping = {
            'RUNNING'    : RUNNING,
            'CANCELLED'  : CANCELLED,
            'COMPLETED'  : COMPLETED,
            'CONFIGURING': PENDING,
            'COMPLETING' : RUNNING,
            'FAILED'     : FAILED,
            'NODE_FAIL'  : FAILED,
            'PENDING'    : PENDING,
            'PREEMPTED'  : FAILED,
            'SUSPENDED'  : PENDING,
            'TIMEOUT'    : FAILED,
    }
    for line in response.split("\n"):
        if line.strip() in slurm_status_mapping:
            return slurm_status_mapping[line.strip()]
    return None

backend = {
    'name': 'slurm',
    'prefix': '#SBATCH',
    'extension' : 'slr',
    'cmd_submit'         : (['sbatch', ],
                            get_job_id),
    'cmd_status_running' : (['squeue', '-h', '-o %T', '-j {job_id}'],
                            get_job_status),
    'cmd_status_finished': (['sacct', '--format=state', '-n', '-j {job_id}'],
                            get_job_status),
    'cmd_cancel'         : ['scancel', '{job_id}', ],
    'translate_options': translate_options,
    'default_opts': {
        'nodes'  : 1,
        'threads': 1,
    },
    'job_vars': {
        'XXX_JOB_ID'     : 'SLURM_JOB_ID',
        'XXX_WORKDIR'    : 'SLURM_SUBMIT_DIR',
        'XXX_HOST'       : 'SLURM_SUBMIT_HOST',
        'XXX_JOB_NAME'   : 'SLURM_JOB_NAME',
        'XXX_ARRAY_INDEX': 'SLURM_ARRAY_TASK_ID',
        'XXX_NODELIST'   : 'SLURM_JOB_NODELIST',
    },
}



"""Generalized status codes for submitted jobs"""
PENDING   = -2
RUNNING   = -1
COMPLETED =  0
CANCELLED =  1
FAILED    =  2

STATUS_CODES = [PENDING, RUNNING, COMPLETED, CANCELLED, FAILED]

str_status = {
 PENDING   : 'PENDING',
 RUNNING   : 'RUNNING',
 COMPLETED : 'COMPLETED',
 CANCELLED : 'CANCELLED',
 FAILED    : 'FAILED',
}




"""
Collection of utility functions
"""
import os
import stat
import subprocess as sp


def set_executable(filename):
    """Set the execuable bit on the given filename"""
    st = os.stat(filename)
    os.chmod(filename, st.st_mode | stat.S_IEXEC)


def write_file(filename, data):
    """Write data to the file with the given filename"""
    with open(filename, 'w') as out_fh:
        out_fh.write(data)


def read_file(filename):
    """
    Return the contents of the file with the given filename as a string

    >>> write_file('read_write_file.txt', 'Hello World')
    >>> read_file('read_write_file.txt')
    'Hello World'
    >>> os.unlink('read_write_file.txt')
    """
    with open(filename) as in_fh:
        return in_fh.read()


def run_cmd(cmd, remote, workdir=None, ignore_exit_code=False):
    r'''
    Run the given cmd in the given workdir, either locally or remotely, and
    return the combined stdout/stderr

    Parameters
    ----------

    cmd: list of str
        Command to execute, as list of command and options

    remote: None or str
        If None, run command locally. Otherwise, run on the given host (via
        SSH)

    workdir: str, optional
        Local or remote directory from which to run the command.

    ignore_exit_code: boolean, optional
        By default, subprocess.CalledProcessError will be raised if the call
        has an exit code other than 0. This exception can be supressed by
        passing `ignore_exit_code=False`

    Example
    -------

    >>> import tempfile, os, shutil
    >>> tempfolder = tempfile.mkdtemp()
    >>> scriptfile = os.path.join(tempfolder, 'test.sh')
    >>> with open(scriptfile, 'w') as script_fh:
    ...     script_fh.writelines(["#!/bin/bash\n", "echo Hello $1\n"])
    >>> set_executable(scriptfile)

    >>> run_cmd(['./test.sh', 'World'], remote=None, workdir=tempfolder)
    'Hello World\n'

    >>> shutil.rmtree(tempfolder)
    '''
    assert type(cmd) in [list, tuple], "cmd must be given as a list"
    try:
        if remote is None: # run locally
            if workdir is None:
                response = sp.check_output(cmd, stderr=sp.STDOUT)
            else:
                response = sp.check_output(cmd, stderr=sp.STDOUT, cwd=workdir)
        else: # run remotely
            cmd = " ".join(cmd)
            if workdir is None:
                cmd = ['ssh', remote, cmd]
            else:
                cmd = ['ssh', remote, 'cd %s && %s' % (workdir, cmd)]
            response = sp.check_output(cmd, stderr=sp.STDOUT)
    except sp.CalledProcessError as e:
        if ignore_exit_code:
            response = e.output
        else:
            raise
    return response


def time_to_seconds(time_str):
    """
    Convert a string describing a time duration into seconds. The supported
    formats are: "minutes", "minutes:seconds", "hours:minutes:seconds",
    "days-hours", "days-hours:minutes" and "days-hours:minutes:seconds"

    >>> time_to_seconds('10')
    600
    >>> time_to_seconds('10:00')
    600
    >>> time_to_seconds('10:30')
    630
    >>> time_to_seconds('1:10:30')
    4230
    >>> time_to_seconds('1-1:10:30')
    90630
    >>> time_to_seconds('1-0')
    86400
    >>> time_to_seconds('1-10')
    122400
    >>> time_to_seconds('1-1:10')
    90600
    >>> time_to_seconds('1-1:10:30')
    90630
    >>> time_to_seconds('1 1:10:30')
    Traceback (most recent call last):
      ...
    ValueError: '1 1:10:30' has invalid pattern
    """
    import re
    patterns = [
        re.compile(r'^(?P<hours>\d+):(?P<minutes>\d+):(?P<seconds>\d+)$'),
        re.compile(r'^(?P<days>\d+)-(?P<hours>\d+)$'),
        re.compile(r'^(?P<minutes>\d+)$'),
        re.compile(r'^(?P<minutes>\d+):(?P<seconds>\d+)$'),
        re.compile(r'^(?P<days>\d+)-(?P<hours>\d+):(?P<minutes>\d+)$'),
        re.compile(
          r'^(?P<days>\d+)-(?P<hours>\d+):(?P<minutes>\d+):(?P<seconds>\d+)$'),
    ]
    seconds = 0
    for pattern in patterns:
        match = pattern.match(str(time_str).strip())
        if match:
            if 'seconds' in match.groupdict():
                seconds += int(match.group('seconds'))
            if 'minutes' in match.groupdict():
                seconds += 60*int(match.group('minutes'))
            if 'hours' in match.groupdict():
                seconds += 3600*int(match.group('hours'))
            if 'days' in match.groupdict():
                seconds += 86400*int(match.group('days'))
            return seconds
    raise ValueError("'%s' has invalid pattern" % time_str)


def mkdir(name, mode=0750):
    """
    Implementation of 'mkdir -p': Creates folder with the given `name` and the
    given permissions (`mode`)

    * Create missing parents folder
    * Do nothing if the folder with the given `name` already exists
    * Raise OSError if there is already a file with the given `name`
    """
    if os.path.isdir(name):
        pass
    elif os.path.isfile(name):
        raise OSError("A file with the same name as the desired " \
                      "dir, '%s', already exists." % name)
    else:
        os.makedirs(name, mode)
